import numpy as np
import pandas as pd

from sklearn.model_selection import KFold, cross_val_score
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.metrics import mean_squared_error

from sklearn.cluster import KMeans
from sklearn.model_selection import GroupKFold

from pykrige.rk import RegressionKriging

from sdswrapper.ordinarykriginginterface import OrdinaryKrigingInterface



class Models:
    """
    Class for training and evaluating machine learning and kriging models.

    Attributes:
        models (list): List of models to be trained.
        X (pd.DataFrame): Input data for training.
        P (pd.DataFrame): Spatial projection data.
        XP (pd.DataFrame): Combination of X and P.
        Y (pd.DataFrame): Output (target) data.
        k (int): Number of folds for cross-validation.
        projections_folder (str): Path to save projections.
        spatial_groups (np.ndarray): Spatial groups for cross-validation.
    """

    def __init__(self, models: list, X: pd.DataFrame, p: pd.DataFrame, y: pd.DataFrame, k: int, projections_folder: str) -> None:
        """
        Initializes the Models class with the provided data and settings.

        Args:
            models (list): List of models to be trained.
            X (pd.DataFrame): Input data for training.
            p (pd.DataFrame): Spatial projection data.
            y (pd.DataFrame): Output (target) data.
            k (int): Number of folds for cross-validation.
            projections_folder (str): Path to save projections.

        Raises:
            Exception: If the data contains infinite or NaN values.
        """

        if np.isinf(X).any().any():

            raise Exception("`X` contains infinite values.")

        if np.isinf(y).any():

            raise Exception("`y` contains infinite values.")

        if np.isinf(p).any().any():

            raise Exception("`p` contains infinite values.")

        if np.isnan(X).any().any():

            raise Exception("`X` contains infinite values.")

        if np.isnan(y).any():

            raise Exception("`y` contains infinite values.")

        if np.isnan(p).any().any():

            raise Exception("`p` contains infinite values.")


        self.models = models
        self.X = self.set_X(X)
        self.P = self.set_P(p)
        self.XP = self.set_XP(self.X, self.P)
        self.Y = self.set_Y(y)
        self.k = k
        self.sample_size = X.shape[0]
        self.projections_folder = projections_folder
        self.spatial_groups = KMeans(n_clusters=self.k, random_state=42).fit_predict(self.P)



    def __check_data(self, data: pd.DataFrame):
        """
        Checks if the data contains NaN or infinite values.

        Args:
            data (pd.DataFrame): Data to be checked.

        Returns:
            pd.DataFrame: Verified data.

        Raises:
            Exception: If the data contains NaN or infinite values.
        """

        # NaN
        if np.isnan(data).any().any():

            print()
            print('*data = ', data)
            print()

            raise Exception("`data` contains NaN values.")

        # Inf
        if np.isinf(data).any().any():

            print()
            print('*data = ', data)
            print()

            raise Exception("`data` contains infinite values.")

        return data


    def set_X(self, X: pd.DataFrame):
        """
        Sets the input data X after verification and conversion.

        Args:
            X (pd.DataFrame): Input data.

        Returns:
            pd.DataFrame: Processed input data.
        """

        X = X.reset_index(drop=True).astype('float32')

        X = self.__check_data(X)

        return X


    def set_P(self, p: pd.DataFrame):
        """
        Sets the projection data P after verification and conversion.

        Args:
            p (pd.DataFrame): Projection data.

        Returns:
            pd.DataFrame: Processed projection data.
        """

        p = p.reset_index(drop=True).astype('float32')

        P = self.__check_data(p)

        return p


    def set_Y(self, y: pd.DataFrame):
        """
        Sets the output data Y after verification and conversion.

        Args:
            y (pd.DataFrame): Output data.

        Returns:
            pd.DataFrame: Processed output data.
        """

        y = y.reset_index(drop=True).astype('float32')

        y = self.__check_data(y)

        return y


    def set_XP(self, X: pd.DataFrame, p: pd.DataFrame):
        """
        Combines the input data X and projection data P.

        Args:
            X (pd.DataFrame): Input data.
            p (pd.DataFrame): Projection data.

        Returns:
            pd.DataFrame: Combination of X and P.
        """

        return pd.concat([X, p], axis=1).astype('float32').copy()


    def fit(self):
        """
        Trains the provided models using cross-validation.

        Returns:
            list: List of dictionaries containing metrics and trained models.
        """

        output = list()

        for name, model in self.models:

            model_type = None

            if isinstance(model, RegressionKriging):

                model_type = 'KR'

                model_metrics, model_trained = self._fit_regression_kriging(model)

            elif isinstance(model, OrdinaryKrigingInterface):

                model_type = 'KR'

                model_metrics, model_trained = self._fit_ordinary_kriging_models(model)

            else:

                model_type = 'SK'

                model_metrics, model_trained = self._fit_sklearn_models(model)

            output.append(
                {
                    'sample_size': self.sample_size, 
                    'name': name, 
                    'model_type': model_type, 
                    'model_metrics_mean': model_metrics.mean(), 
                    'model_metrics_std': model_metrics.std(),
                    'trained_model': model_trained
                 }
            )

        return output


    def _fit_sklearn_models(self, model):
        """
        Trains scikit-learn models using spatial cross-validation.

        Args:
            model: Scikit-learn model to be trained.

        Returns:
            tuple: RMSE metrics and trained model.
        """

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            ('regressor', model)
        ])

        # TODO: ajustar esta Validação cruzada para spatial cross-validation
        # kfold = KFold(n_splits=self.k, shuffle=True, random_state=42)
        group_kfold = GroupKFold(n_splits=self.k)

        scores = cross_val_score(pipeline, 
                                 X=self.P, y=self.Y,
                                 scoring='neg_mean_squared_error', 
                                 cv=group_kfold.split(self.P, self.Y, groups=self.spatial_groups))

        mse_scores = -scores 
        rmse_scores = np.sqrt(mse_scores)


        model.fit(self.P.values, self.Y.values)

        return rmse_scores, model


    def _fit_ordinary_kriging_models(self, model):
        """
        Trains ordinary kriging models using spatial cross-validation.

        Args:
            model: Ordinary kriging model to be trained.

        Returns:
            tuple: RMSE metrics and trained model.
        """

        rmse_scores = list()
        group_kfold = GroupKFold(n_splits=self.k)

        # TODO: ajustar esta Validação cruzada para spatial cross-validation
        # kf = KFold(n_splits=self.k, shuffle=True, random_state=42)

        # kf_idxs = [x for x in kf.split(self.Y.index)]

        for train_idx, test_idx in group_kfold.split(self.P, self.Y, groups=self.spatial_groups):
        # for k in range(self.k):

            X_train, X_test = self.X.iloc[train_idx].copy(), self.X.iloc[test_idx].copy()
            Y_train, Y_test = self.Y.iloc[train_idx].copy(), self.Y.iloc[test_idx].copy()

            # X_train = self.X.loc[kf_idxs[k][0]].copy()
            # X_test = self.X.loc[kf_idxs[k][1]].copy()

            # Y_train = self.Y.loc[kf_idxs[k][0]].copy()
            # Y_test = self.Y.loc[kf_idxs[k][1]].copy()


            X_train = X_train.values
            X_test = X_test.values

            Y_train = Y_train.values
            Y_test = Y_test.values


            model.fit(
                X = X_train,
                y = Y_train
            )


            predictions = model.predict(
                X = X_test
            )

            predictions = predictions.astype(np.float32)


            mse = mean_squared_error(Y_test, predictions)

            rmse = np.sqrt(mse)

            rmse_scores.append(rmse)


        model.fit(X = self.X.values, y = self.Y.values)

        return np.array(rmse_scores), model


    def _fit_regression_kriging(self, model):
        """
        Trains regression kriging models using spatial cross-validation.

        Args:
            model: Regression kriging model to be trained.

        Returns:
            tuple: RMSE metrics and trained model.
        """

        rmse_scores = list()
        group_kfold = GroupKFold(n_splits=self.k)

        # TODO: ajustar esta Validação cruzada para spatial cross-validation
        # kf = KFold(n_splits=self.k, shuffle=True, random_state=42)

        # kf_idxs = [x for x in kf.split(self.Y.index)]

        for train_idx, test_idx in group_kfold.split(self.P, self.Y, groups=self.spatial_groups):
        # for k in range(self.k):

            # X_train = self.X.loc[kf_idxs[k][0]].copy()
            # X_test = self.X.loc[kf_idxs[k][1]].copy()

            # P_train = self.P.loc[kf_idxs[k][0]].copy()
            # P_test = self.P.loc[kf_idxs[k][1]].copy()

            # Y_train = self.Y.loc[kf_idxs[k][0]].copy()
            # Y_test = self.Y.loc[kf_idxs[k][1]].copy()

            X_train, X_test = self.X.iloc[train_idx].copy(), self.X.iloc[test_idx].copy()
            P_train, P_test = self.P.iloc[train_idx].copy(), self.P.iloc[test_idx].copy()
            Y_train, Y_test = self.Y.iloc[train_idx].copy(), self.Y.iloc[test_idx].copy()

            X_train = X_train.values
            X_test = X_test.values

            P_train = P_train.values
            P_test = P_test.values

            Y_train = Y_train.values
            Y_test = Y_test.values


            Y_train = np.where(np.isinf(Y_train), np.nan, Y_train)
            Y_test  = np.where(np.isinf(Y_test),  np.nan, Y_test)

            Y_train = np.where(np.isnan(Y_train), np.nanmean(Y_train), Y_train)
            Y_test  = np.where(np.isnan(Y_test),  np.nanmean(Y_test),  Y_test)


            model.fit(
                p = P_train,
                x = X_train,
                y = Y_train
            )

            predictions = model.predict(
                p = P_test,
                x = X_test
            )

            predictions = predictions.astype(np.float32)


            mse = mean_squared_error(Y_test, predictions)

            rmse = np.sqrt(mse)

            rmse_scores.append(rmse)


        model.fit(p = self.P.values, x = self.X.values, y = self.Y.values)

        rmse_scores = np.array(rmse_scores)

        return (rmse_scores, model)


    # def projection(self, X_full:pd.DataFrame, P_full:pd.DataFrame,
    #                shape:tuple, tag:str = None):


    #     if not isinstance(X_full, pd.DataFrame):

    #         raise Exception("`X_full` must be pandas DataFrame.")

    #     if not isinstance(P_full, pd.DataFrame):

    #         raise Exception("`P_full` must be pandas DataFrame.")

    #     # if not isinstance(Y_full, pd.Series):

    #     #     raise Exception("`Y_full` must be pandas DataFrame.")

    #     if not isinstance(shape, tuple):

    #         raise Exception("`shape` must be a tuple.")

    #     if (tag != None) and (not isinstance(tag, str)):

    #         raise Exception("`shape` must be a tuple.")

    #     if np.isinf(X_full).any().any():

    #         raise Exception("`X_full` contains infinite values.")

    #     # if np.isinf(Y_full).any():

    #     #     raise Exception("`Y_full` contains infinite values.")

    #     if np.isinf(P_full).any().any():

    #         raise Exception("`P_full` contains infinite values.")

    #     if np.isnan(X_full).any().any():

    #         raise Exception("`X_full` contains infinite values.")

    #     # if np.isnan(Y_full).any():

    #     #     raise Exception("`Y_full` contains infinite values.")

    #     if np.isnan(P_full).any().any():

    #         raise Exception("`P_full` contains infinite values.")


    #     X_full_raw = X_full.copy()
    #     P_full_raw = P_full.copy()
    #     # Y_full_raw = Y_full.copy()


    #     ################# TODO: apagar depois
    #     # print()
    #     # print('~~'*20)
    #     # print('# PARTE B')
    #     # print('type(X_full):', type(X_full))
    #     # print('type(P_full):', type(P_full))
    #     # print('type(Y_full):', type(Y_full))
    #     # print()
    #     # print('~~'*20)
    #     # print()
    #     #################


    #     if not os.path.exists(self.projections_folder):

    #         os.mkdir(self.projections_folder)


    #     for name, model in self.models:

    #         print('# Model:', name) # TODO: comentar esta linha depois

    #         model_type = None


    #         # Regression Kriging
    #         if isinstance(model, RegressionKriging):

    #             model_type = 'KR'

    #             # X = df_fulldata[X_column_names]
    #             # p = df_fulldata[P_colum_names]
    #             # y = df_fulldata[Y_column_name]

    #             ################# TODO: apagar depois
    #             # print()
    #             # print('~~'*20)
    #             # print(f'# PARTE C em cima - {name}')
    #             # print('type(X_full):', type(X_full))
    #             # print('type(P_full):', type(P_full))
    #             # print('type(Y_full):', type(Y_full))
    #             # print()
    #             # print('~~'*20)
    #             # print()
    #             #################

    #             y_data = self.Y.values.copy()
    #             y_data = np.where(np.isinf(y_data), np.nan, y_data)
    #             y_data = np.where(np.isnan(y_data), np.nanmean(y_data), y_data)

    #             model.fit(
    #                 x = self.X.values.copy(),
    #                 p = self.P.values.copy(),
    #                 y = y_data
    #             )


    #             X_full = X_full_raw.copy()
    #             X_full = X_full.reset_index(drop=True).copy()
    #             X_full = X_full.values
    #             X_full = np.where(np.isinf(X_full), np.nan, X_full)
    #             X_full = np.where(np.isnan(X_full), np.nanmean(X_full), X_full)

    #             P_full = P_full_raw.copy()
    #             P_full = P_full.reset_index(drop=True).copy()
    #             P_full = P_full.values
    #             P_full = np.where(np.isinf(P_full), np.nan, P_full)
    #             P_full = np.where(np.isnan(P_full), np.nanmean(P_full), P_full)


    #             ################# TODO: apagar depois
    #             # print()
    #             # print('~~'*20)
    #             # print(f'# PARTE C embaixo - {name}')
    #             # print('type(X_full):', type(X_full))
    #             # print('type(P_full):', type(P_full))
    #             # print('type(Y_full):', type(Y_full))
    #             # print()
    #             # print('~~'*20)
    #             # print()
    #             #################


    #             projection = model.predict(
    #                 x = X_full,
    #                 p = P_full
    #             )

    #             projection = projection.astype(np.float32)
    #             projection = projection.reshape(shape)

    #             projection = np.where(np.isinf(projection), np.nan, projection)
    #             projection = np.where(np.isnan(projection), np.nanmean(projection), projection)

    #             hash = datetime.now().strftime("%Y%m%d_%H%M")
    #             hash = tag + '_' + hash if tag != None else hash

    #             with open(os.path.join(self.projections_folder, f'{name}_projection_{hash}.pkl'), 'wb') as f:

    #                 pickle.dump(projection, f)


    #         # Ordinary Kriging
    #         elif isinstance(model, OrdinaryKrigingInterface):

    #             ################# TODO: apagar depois
    #             # print()
    #             # print('~~'*20)
    #             # print(f'# PARTE D - elif - {name}')
    #             # print('type(X_full):', type(X_full))
    #             # print('type(P_full):', type(P_full))
    #             # print('type(Y_full):', type(Y_full))
    #             # print()
    #             # print('~~'*20)
    #             # print()
    #             #################

    #             model_type = 'KR'

    #             y_data = self.Y.values.copy()
    #             y_data = np.where(np.isinf(y_data), np.nan, y_data)
    #             y_data = np.where(np.isnan(y_data), np.nanmean(y_data), y_data)

    #             model.fit(
    #                 self.X.values.copy(),
    #                 y_data
    #             )

    #             X_full = X_full_raw.copy()
    #             X_full = X_full.reset_index(drop=True).copy()
    #             X_full = X_full.values
    #             X_full = np.where(np.isinf(X_full), np.nan, X_full)
    #             X_full = np.where(np.isnan(X_full), np.nanmean(X_full), X_full)

    #             projection = model.predict(
    #                 X_full
    #             )

    #             projection = projection.astype(np.float32)
    #             projection = projection.reshape(shape)

    #             projection = np.where(np.isinf(projection), np.nan, projection)
    #             projection = np.where(np.isnan(projection), np.nanmean(projection), projection)

    #             hash = datetime.now().strftime("%Y%m%d_%H%M")
    #             hash = tag + '_' + hash if tag != None else hash

    #             with open(os.path.join(self.projections_folder, f'{name}_projection_{hash}.pkl'), 'wb') as f:

    #                 pickle.dump(projection, f)

    #         else:

    #             ################# TODO: apagar depois
    #             # print()
    #             # print('~~'*20)
    #             # print(f'# PARTE Ea - else - {name}')
    #             # print('type(X_full):', type(X_full))
    #             # print('type(P_full):', type(P_full))
    #             # print('type(Y_full):', type(Y_full))
    #             # print()
    #             # print('~~'*20)
    #             # print()
    #             #################


    #             model_type = 'SK'

    #             # X = df_fulldata[X_column_names + P_colum_names]
    #             # y = df_fulldata[Y_column_name]

    #             # model.fit(self.XP, self.Y)
    #             model.fit(
    #                 self.P.values.copy(),
    #                 self.Y.values.copy()
    #             )

    #             ################# TODO: apagar depois
    #             # print()
    #             # print('~~'*20)
    #             # print(f'# PARTE Eb - else - {name}')
    #             # print('type(X_full):', type(X_full))
    #             # print('type(P_full):', type(P_full))
    #             # print('type(Y_full):', type(Y_full))
    #             # print()
    #             # print('~~'*20)
    #             # print()
    #             #################

    #             # X_full.reset_index(drop=True, inplace=True)
    #             # P_full = P_full#.reset_index(drop=True).copy()
    #             # XP = pd.concat([X_full, P_full], axis=1)

    #             ################ TODO: apagar depois
    #             print()
    #             print('~~'*20)
    #             print(f'# PARTE Ec - else - {name}')
    #             print('type(X_full):', type(X_full))
    #             print('type(P_full):', type(P_full))
    #             # print('type(Y_full):', type(Y_full))
    #             print()
    #             print('~~'*20)
    #             print()
    #             ################

    #             P_full = P_full_raw.copy()
    #             P_full = P_full.astype('float32') #.astype('float32').values.astype(np.float32)

    #             ################ TODO: apagar depois
    #             print()
    #             print('~~'*20)
    #             print(f'# PARTE Ec ALPHA - else - {name}')
    #             print('type(X_full):', type(X_full))
    #             print('type(P_full):', type(P_full))
    #             # print('type(Y_full):', type(Y_full))
    #             print()
    #             print('~~'*20)
    #             print()
    #             ################

    #             P_full = np.where(np.isinf(P_full), np.nan, P_full)

    #             ################# TODO: apagar depois
    #             print()
    #             print('~~'*20)
    #             print(f'# PARTE Ec BETHA - else - {name}')
    #             print('type(X_full):', type(X_full))
    #             print('type(P_full):', type(P_full))
    #             # print('type(Y_full):', type(Y_full))
    #             print()
    #             print('~~'*20)
    #             print()
    #             #################

    #             P_full = np.where(np.isnan(P_full), np.nanmean(P_full), P_full)


    #             ################# TODO: apagar depois
    #             print()
    #             print('~~'*20)
    #             print(f'# PARTE Ed - else - {name}')
    #             print('type(X_full):', type(X_full))
    #             print('type(P_full):', type(P_full))
    #             # print('type(Y_full):', type(Y_full))
    #             print()
    #             print('~~'*20)
    #             print()
    #             #################

    #             # projection = model.predict(XP)
    #             projection = model.predict(
    #                 P_full#.values
    #             )

    #             projection = projection.astype(np.float32)
    #             projection = projection.reshape(shape)

    #             projection = np.where(np.isinf(projection), np.nan, projection)
    #             projection = np.where(np.isnan(projection), np.nanmean(projection), projection)

    #             hash = datetime.now().strftime("%Y%m%d_%H%M")
    #             hash = tag + '_' + hash if tag != None else hash

    #             with open(os.path.join(self.projections_folder, f'{name}_projection_{hash}.pkl'), 'wb') as f:

    #                 pickle.dump(projection, f)



import os
import pickle

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from sdswrapper.sampler import SampleGenerator
from sdswrapper.models import Models

from sklearn.metrics import mean_squared_error

from rasterio.mask import mask
from rasterio.plot import show

from pykrige.rk import RegressionKriging
from pykrige.ok import OrdinaryKriging
from sdswrapper.ordinarykriginginterface import OrdinaryKrigingInterface



class Wrapper:
    """
    Wrapper class for managing model training, prediction, and evaluation.

    Attributes:
        model_list (list): List of models to be trained.
        dataset (pd.DataFrame): Dataset containing input and output data.
        X_colum_names (str): Names of input columns.
        p_colum_names (str): Names of projection columns.
        y_column_name (str): Name of the output column.
        projections_folder (str): Path to save projections.
        k (int): Number of folds for cross-validation.
        gridsearch_parameters (list): List of tuples for grid search parameters.
    """

    def __init__(self, 
                 model_list: list,
                 dataset: pd.DataFrame,
                 X_column_names: str,
                 P_column_names: str,
                 y_column_name: str,
                 projections_folder: str,
                 k: int,
                 gridsearch_parameters: list = None):
        """
        Initializes the Wrapper class with the provided parameters.

        Args:
            model_list (list): List of models to be trained.
            dataset (pd.DataFrame): Dataset containing input and output data.
            X_column_names (str): Names of input columns.
            P_column_names (str): Names of projection columns.
            y_column_name (str): Name of the output column.
            projections_folder (str): Path to save projections.
            k (int): Number of folds for cross-validation.
            gridsearch_parameters (list, optional): List of tuples for grid search parameters.
        """
        self.model_list = model_list
        self.dataset = dataset
        self.X_colum_names = X_column_names
        self.p_colum_names = P_column_names
        self.y_column_name = y_column_name
        self.gridsearch_parameters = gridsearch_parameters
        # self.projector = projector # TODO
        self.k = k
        # self.tag = tag
        self.projections_folder = projections_folder


    def fit(self):
        """
        Trains the models using the provided dataset and parameters.

        Returns:
            dict: Profiles of trained models.
        """
        model_profiles = dict()

        try:

            data = self.__check_data(self.dataset)

        except Exception as e:

            raise Exception(e)


        X = data[self.X_colum_names].astype(float).copy()
        p = data[self.p_colum_names].astype(float).copy()
        y = data[self.y_column_name].astype(float).copy()

        try:

            models = self.set_models(X, p, y)

            model_profiles = models.fit()

        except Exception as e:

            raise Exception(e)


        return model_profiles


    def set_models(self, X: pd.DataFrame, p: pd.DataFrame, y: pd.DataFrame):
        """
        Configures the models with the provided data.

        Args:
            X (pd.DataFrame): Input data.
            p (pd.DataFrame): Projection data.
            y (pd.DataFrame): Output data.

        Returns:
            Models: Configured models instance.
        """
        return Models(
            models = self.model_list,
            X = X,
            p = p,
            y = y,
            k = self.k,
            projections_folder = self.projections_folder
        )


    def __check_data(self, data: pd.DataFrame, numpy_array: bool = False):
        """
        Validates the dataset for NaN or infinite values.

        Args:
            data (pd.DataFrame): Dataset to validate.
            numpy_array (bool): Whether the data is a numpy array.

        Returns:
            pd.DataFrame: Validated dataset.

        Raises:
            Exception: If the dataset contains NaN or infinite values.
        """
        if numpy_array:

            # NaN
            if np.isnan(data).any().any():

                raise Exception("`data` contains NaN values.")

            # Inf
            if np.isinf(data).any().any():

                raise Exception("`data` contains infinite values.")

            return data

        else:

            # NaN
            if np.isnan(data).values.any().any():

                raise Exception("`data` contains NaN values.")

            # Inf
            if np.isinf(data).values.any().any():

                raise Exception("`data` contains infinite values.")

            return data


    @staticmethod
    def predict(trained_model: Models | list, X: pd.DataFrame, 
                p: pd.DataFrame = None, shape: tuple = None):
        """
        Makes predictions using the trained models.

        Args:
            trained_model (Models | list): Trained model or list of models.
            X (pd.DataFrame): Input data for prediction.
            p (pd.DataFrame, optional): Projection data for prediction.
            shape (tuple): Shape of the output prediction.

        Returns:
            np.ndarray: Predictions.

        Raises:
            TypeError: If input types are incorrect.
        """
        if not isinstance(shape, tuple):

            raise TypeError('`shape` must be a tuple.')

        if not isinstance(X, pd.DataFrame):

            raise TypeError('`X` must be a pandas DataFrame.')

        if not isinstance(p, pd.DataFrame):

            raise TypeError('`p` must be a pandas DataFrame.')


        X = X.astype(float).values.copy()
        p = p.astype(float).values.copy()


        if not isinstance(trained_model, list):

            prediction = Wrapper._predict_single_model(trained_model, X, p, shape)

        elif isinstance(trained_model, list):

            prediction = Wrapper._predict_multiple_models(trained_model, X, p, shape)

        else:

            raise TypeError('`trained_model` must be a RegressionKriging, OrdinaryKrigingInterface, sklearn model instance, or a python list instance.')
        
        return prediction


    @staticmethod
    def _predict_multiple_models(trained_models: list, X: np.ndarray, 
                                 p: np.ndarray = None, shape: tuple = None):
        """
        Makes predictions using multiple trained models.

        Args:
            trained_models (list): List of trained models.
            X (np.ndarray): Input data for prediction.
            p (np.ndarray, optional): Projection data for prediction.
            shape (tuple): Shape of the output prediction.

        Returns:
            np.ndarray: Predictions from multiple models.

        Raises:
            TypeError: If input types are incorrect.
        """
        if not isinstance(shape, tuple):

            raise TypeError('`shape` must be a tuple.')

        if not isinstance(X, np.ndarray):

            raise TypeError('`X` must be a numpy array.')

        if not isinstance(p, np.ndarray):

            raise TypeError('`p` must be a numpy array.')

        if not isinstance(trained_models, list):

            raise TypeError('`trained_models` must be a python list.')

        predictions = []

        for model in trained_models:

            prediction = Wrapper._predict_single_model(model['trained_model'], X, p, shape)

            predictions.append(prediction)

        return np.array(predictions)


    @staticmethod
    def _predict_single_model(model: Models, X: np.ndarray, 
                              p: np.ndarray = None, shape: tuple = None):
        """
        Makes predictions using a single trained model.

        Args:
            model (Models): Trained model.
            X (np.ndarray): Input data for prediction.
            p (np.ndarray, optional): Projection data for prediction.
            shape (tuple): Shape of the output prediction.

        Returns:
            np.ndarray: Predictions from the model.

        Raises:
            TypeError: If input types are incorrect.
        """
        if not isinstance(shape, tuple):

            raise TypeError('`shape` must be a tuple.')

        if not isinstance(X, np.ndarray):

            raise TypeError('`X` must be a numpy array.')

        if not isinstance(p, np.ndarray):

            raise TypeError('`p` must be a numpy array.')


        if isinstance(model, RegressionKriging):

            prediction = model.predict(X, p)

        elif isinstance(model, OrdinaryKrigingInterface):

            prediction = model.predict(X)

        else:

            try:

                prediction = model.predict(p)

            except:

                raise TypeError('`model` must be a RegressionKriging, OrdinaryKrigingInterface or a sklearn model instance.')

        return prediction.reshape(shape)


    @staticmethod
    def mask(prediction: np.ndarray, nan_mask: np.ndarray):
        """
        Applies a mask to the prediction to handle NaN values.

        Args:
            prediction (np.ndarray): Prediction data.
            nan_mask (np.ndarray): Mask to apply.

        Returns:
            np.ndarray: Masked prediction.

        Raises:
            TypeError: If input types are incorrect.
            ValueError: If shapes of prediction and mask do not match.
        """
        if not isinstance(prediction, np.ndarray):

            raise TypeError('`prediction` must be a numpy array.')

        if not isinstance(nan_mask, np.ndarray):

            raise TypeError('`nan_mask` must be a numpy array.')
        
        if prediction.shape != nan_mask.shape:

            raise ValueError('`prediction` and `nan_mask` must have the same shape.')


        masked_prediction = np.where(np.isnan(nan_mask), np.nan, prediction)


        return masked_prediction


    @staticmethod
    def score(prediction: np.ndarray, reference_data: np.ndarray):
        """
        Calculates the RMSE score between prediction and reference data.

        Args:
            prediction (np.ndarray): Predicted data.
            reference_data (np.ndarray): Reference data.

        Returns:
            float: RMSE score.

        Raises:
            TypeError: If input types are incorrect.
        """
        if not isinstance(prediction, np.ndarray):

            raise TypeError('`prediction` must be a numpy array.')

        if not isinstance(reference_data, np.ndarray):

            raise TypeError('`reference_data` must be a numpy array.')

        mse = mean_squared_error(
            prediction.flatten(),
            reference_data.flatten()
        )

        return float(np.sqrt(mse))


    @staticmethod
    def plot(predictions: pd.DataFrame, save: bool = False, folder: str = None, best: bool = True, **kwargs):
        """
        Plots the predictions and optionally saves the plots.

        Args:
            predictions (pd.DataFrame): DataFrame containing predictions.
            save (bool): Whether to save the plots.
            folder (str, optional): Folder to save the plots.
            best (bool): Whether to plot only the best prediction.
            **kwargs: Additional arguments for customization.

        Raises:
            TypeError: If input types are incorrect.
        """
        column_name = kwargs.get('column_name', None)

        if (column_name is not None) and (not isinstance(column_name, str)):

            raise TypeError('`column_name` must be a string.')

        if not isinstance(predictions, pd.DataFrame):

            raise TypeError('`predictions` must be a pandas DataFrame.')

        if (folder is not None) and  (not isinstance(save, bool)):

            raise TypeError('`save` must be a boolean.')

        if (folder is not None) and (not isinstance(folder, str)):

            raise TypeError('`folder` must be a string.')


        reference_data = kwargs.get('reference_data', None)

        n_cols = 1 if reference_data is None else 3


        if best:

            column_name = 'model_metrics_mean' if column_name is None else column_name

            predictions = predictions.sort_values(by=column_name, ascending=True).iloc[[0]].copy()

            predictions.reset_index(drop=True, inplace=True)

    
        if reference_data is None:

            for i in range(predictions.shape[0]):
                
                prediction = predictions.iloc[i]['prediction']
                name = predictions.iloc[i]['name']

                fig = plt.figure(figsize=(12, 5))

                plt.imshow(prediction, cmap='coolwarm', vmin=0, vmax=1000)
                plt.colorbar()
                plt.title(name) #, fontsize=16)

                # Adicionando pontos do dataset (descomente se desejar)
                # ax1.scatter(
                #     experiment_pipeline.df_full_data['coordenada_X'],
                #     experiment_pipeline.df_full_data['coordenada_Y'],
                #     c='red', s=10, label='Dataset Points'
                # )
                # ax1.legend()

                # plt.tight_layout()

                if save:

                    if not os.path.exists(folder):

                        os.makedirs(folder)

                    plt.savefig(os.path.join(folder, f'{i}_{name}.png'))

                    plt.close()

                else:

                    plt.show()

        else:

            for i in range(predictions.shape[0]):
                
                prediction = predictions.iloc[i]['prediction']
                name = predictions.iloc[i]['name']

                fig = plt.figure(figsize=(12, 5))

                # Mapa observado
                ax1 = plt.subplot(1, n_cols, 1)
                im1 = ax1.imshow(reference_data, cmap='coolwarm', vmin=0, vmax=1000)
                # plt.colorbar(im1, ax=ax1)
                ax1.set_title('Observed') #, fontsize=16)
                
                # Adicionando pontos do dataset (descomente se desejar)
                # ax1.scatter(
                #     experiment_pipeline.df_full_data['coordenada_X'],
                #     experiment_pipeline.df_full_data['coordenada_Y'],
                #     c='red', s=10, label='Dataset Points'
                # )
                # ax1.legend()

                # Mapa previsto
                ax2 = plt.subplot(1, n_cols, 2)
                im2 = ax2.imshow(prediction, origin='lower', cmap='coolwarm', vmin=0, vmax=1000)
                # plt.colorbar(im2, ax=ax2)
                ax2.invert_yaxis()
                ax2.set_title(name) #, fontsize=16)


                # Mapa de diferença
                ax3 = plt.subplot(1, n_cols, 3)
                diff = prediction - reference_data
                im3 = ax3.imshow(diff, origin='lower', cmap='coolwarm', vmin=-1000, vmax=1000)
                plt.colorbar(im3, ax=ax3)
                ax3.invert_yaxis()
                ax3.set_title('Difference') #, fontsize=16)

                plt.tight_layout()
                # plt.show()

                if save:

                    if not os.path.exists(folder):

                        os.makedirs(folder)

                    plt.savefig(os.path.join(folder, f'{i}_{name}.png'))

                    plt.close()

                else:

                    plt.show()



from setuptools import setup, find_packages

setup(
    name='sdswrapper',
    version='0.2.8',
    description='A Python package for spatial data science workflows.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Anderson Aires Eduardo',
    author_email='andersonaed@example.com',
    url='https://github.com/AndersonEduardo/sdswrapper',
    packages=find_packages(include=['sdswrapper', 'sdswrapper.*']),  # Inclui todos os submódulos
    include_package_data=True,  # Garante que arquivos adicionais sejam incluídos
    install_requires=[
        'numpy',
        'pandas',
        'matplotlib',
        'scikit-learn',
        'pykrige'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


# TODO: implementar logging corretamente


def load_asc(file_path):

    with open(file_path, 'r') as f:
        header = {}
        for _ in range(6):  # As 6 primeiras linhas são o cabeçalho
            key, value = f.readline().split()
            header[key] = float(value) if '.' in value else int(value)

        # Lendo os dados da matriz
        data = np.loadtxt(f)

    return header, data


def plot_asc(file_path):

    header, data = load_asc(file_path)

    plt.figure(figsize=(10, 8))
    plt.imshow(data, cmap='terrain', origin='upper')
    plt.colorbar(label="Elevação (m)")
    plt.title("Modelo Digital de Elevação (DEM)")
    plt.xlabel("Colunas")
    plt.ylabel("Linhas")
    plt.show()


def replace_with_nan(data):
    """Substitui -9999 por NaN em um array NumPy."""

    data = np.where(data == -9999, np.nan, data)

    return data


def apply_function_to_matrix(data, func):
    """Applies a function to each element of the matrix.

    Args:
        data: The input matrix (NumPy array).
        func: The function to apply to each element.

    Returns:
        A new matrix with the function applied to each element.
    """

    new_data = np.vectorize(func)(data)

    return new_data


def array_to_dataframe(array:np.array):

    output_values = list()

    for y in range(array.shape[0]):

        for x in range(array.shape[1]):

            output_values.append((x, y, array[y, x]))

    return pd.DataFrame(output_values, columns=['x', 'y', 'value'])
